#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "symbol.h"
#include "as_tree.h"
#include "verify.h"
#include "misc.h"

#define err_msg(fmt, arg...) \
do { \
	errors++; \
	message(1, NULL, 0, fmt, ##arg); \
} while (0)

#define warn_msg(fmt, arg...) \
do { \
	message(0, NULL, 0, fmt, ##arg); \
} while (0)

#define note_msg(fmt, arg...) \
do { \
	message(2, NULL, 0, fmt, ##arg); \
} while (0)

typedef void  (*semantic_check_handler_t)(struct ast_node * );

static void semantic_check_access_node(struct ast_node *node);
static void semantic_check_call_node(struct ast_node *node);

struct {
	node_type_t	node_type;
	semantic_check_handler_t handler;
} semantic_checkers[] = {
	{ NODE_TYPE_ACCESS,	semantic_check_access_node },
	{ NODE_TYPE_CALL,	semantic_check_call_node },
	{ NODE_TYPE_UNKNOWN,	NULL }
};

static struct {
	const char      ch;
	data_type_t     type;
} spec2data[] = {
	{ 'd', DATA_TYPE_INTEGER },
	{ 'i', DATA_TYPE_INTEGER },
	{ 'f', DATA_TYPE_REAL },
	{ 's', DATA_TYPE_STRING },
	{ 'c', DATA_TYPE_CHAR },
	{ '0', DATA_TYPE_UNKNOWN }
};

static unsigned int errors;

static data_type_t get_data_type_by_spec(char ch)
{
	unsigned int i;

	for (i = 0; spec2data[i].type != DATA_TYPE_UNKNOWN; i++) {
		if (spec2data[i].ch == ch)
			return spec2data[i].type;
	}

	return DATA_TYPE_UNKNOWN;
}

static void check_fmt_string(struct ast_node_func_call *call,
			     struct symbol_func *symbol)
{
	struct ast_node_const *_const;
	data_type_t *fmtargs;
	unsigned int nfmtargs, i, j;
	boolean_t ok;
	char *ch;
	/* Format string is the last argument in declaration. If it is dynamic
	 * and is calculated at runtime than nothing to check. */
	if (call->args[symbol->nargs-1]->type != NODE_TYPE_CONST)
		return;

	_const = (struct ast_node_const *)call->args[symbol->nargs-1];

	fmtargs = NULL;
	nfmtargs = 0;
	ch = _const->str;
	/* Gather all string specifiers and convert them into data types. */
	while ((ch = strchr(ch, '%')) != NULL) {
		data_type_t type;

		if (ch[1] == '%') {
			ch += 2;
			continue;
		}

		++ch;
		type = get_data_type_by_spec(*ch);

		if (type == DATA_TYPE_UNKNOWN) {
			err_msg("unknown format string specifier %c", *ch);
			goto free;
		}

		nfmtargs++;
		fmtargs = xrealloc(fmtargs, sizeof(data_type_t)*nfmtargs);
		fmtargs[nfmtargs-1] = type;
	}

	if (nfmtargs < (call->nargs - symbol->nargs)) {
		err_msg("too many arguments for format string");
		goto free;
	} else if (nfmtargs > (call->nargs - symbol->nargs)) {
		err_msg("too few arguments for format string");
		goto free;
	}
	/* Time to check nodes type against string specifiers. */
	for (i = symbol->nargs, j = 0; i < call->nargs; i++, j++) {
		ok = verify_expr_type(call->args[i], fmtargs[j]);
		if (!ok) {
			note_msg("argument %d type not fit format string", i+1);
			goto free;
		}
	}

free:
	if (fmtargs != NULL)
		xfree(fmtargs);
}

static void semantic_check_call_node(struct ast_node *node)
{
	struct ast_node_func_call *call;
	struct symbol_func *symbol;
	struct symbol_arg *arg;
	unsigned int i;
	boolean_t ok;

	call = AST_FUNC_CALL(node);
	symbol = (struct symbol_func *)symbol_table_lookup_all(call->name);
	/* Check arguments data type. The syntaxer add default call argument
	 * nodes and we have at least nargs arguments. */
	for (i = 0; i < symbol->nargs; i++) {
		arg = symbol->args[i];

		ok = verify_expr_type(call->args[i], 
				      SYMBOL(arg)->decl_type);
		if (!ok) {
			note_msg("argument %d data type", i+1);
			return;
		}

		if (arg->is_ref) {
			switch (call->args[i]->type) {
			case NODE_TYPE_REFER:
			case NODE_TYPE_ID:
				break;
			default:
				err_msg("not reference type, argument %d", i+1);
				break;
			}
		}
	}
	/* In case of the format string function, we can have more call arguments
	 * than the symbol arguments number. */
	if (symbol->with_fmt && call->nargs > symbol->nargs) {
		/* Check format string arguments. */
		check_fmt_string(call, symbol);
	} else if (call->nargs > symbol->nargs) {
		err_msg("`%s' to many arguments", call->name);
	}
}

static void semantic_check_access_node(struct ast_node *node)
{
	struct ast_node_access *access;
	struct symbol_array *symbol;
	unsigned int i;
	boolean_t ok;

	access = AST_ACCESS(node);
	symbol = (struct symbol_array *)symbol_table_lookup_all(access->name);
	/* When index an array use only integer indexes. */
	for (i = 0; i < access->nidxs; i++) {
		ok = verify_expr_type(access->idxs[i], DATA_TYPE_INTEGER);
		if (!ok)
			note_msg("index data type");	
	}
	/* Check declaration dimenstion against access node use. */
	if (symbol->ndims != access->nidxs)
		err_msg("array was declared with another size");
}

static void semantic_check(struct ast_node *node)
{
	unsigned int i;

	for (i = 0; semantic_checkers[i].node_type != NODE_TYPE_UNKNOWN; i++) {
		if (node->type == semantic_checkers[i].node_type) {
			semantic_checkers[i].handler(node);
			return;
		}
	}

	if (node->child != NULL)
		semantic_check(node->child);

	if (node->next != NULL)
		semantic_check(node->next);
}

unsigned int semantic_analysis(struct ast_node_root *root)
{
	return_val_if_fail(root != NULL, 0);

	semantic_check(AST_NODE(root));

	return errors;
}

